"""
Tests for `pyinaturalist` module.
"""
import json
import os
from datetime import datetime, timedelta

import pytest
import requests_mock
from requests import HTTPError

from pyinaturalist.node_api import get_observation
from pyinaturalist.rest_api import get_access_token, get_all_observation_fields, \
    get_observation_fields, update_observation, create_observations, put_observation_field_values
from pyinaturalist.exceptions import AuthenticationError, ObservationNotFound


def _sample_data_path(filename):
    return os.path.join(os.path.dirname(__file__), 'sample_data', filename)

def _load_sample_json(filename):
    with open(_sample_data_path(filename), encoding='utf-8') as fh:
        return json.load(fh)

PAGE_1_JSON_RESPONSE = _load_sample_json('get_observation_fields_page1.json')
PAGE_2_JSON_RESPONSE = _load_sample_json('get_observation_fields_page2.json')

class TestNodeApi(object):
    @requests_mock.Mocker(kw='mock')
    def test_get_observation(self, **kwargs):
        mock = kwargs['mock']
        mock.get('https://api.inaturalist.org/v1/observations?id=16227955',
                 json=_load_sample_json('get_observation.json'), status_code=200)

        obs_data = get_observation(observation_id=16227955)
        assert obs_data['quality_grade'] == 'research'
        assert obs_data['id'] == 16227955
        assert obs_data['user']['login'] == 'niconoe'
        assert len(obs_data['photos']) == 2

    @requests_mock.Mocker(kw='mock')
    def test_get_non_existent_observation(self, **kwargs):
        mock = kwargs['mock']
        mock.get('https://api.inaturalist.org/v1/observations?id=99999999',
                 json=_load_sample_json('get_nonexistent_observation.json'), status_code=200)
        with pytest.raises(ObservationNotFound):
            get_observation(observation_id=99999999)

class TestRestApi(object):
    @requests_mock.Mocker(kw='mock')
    def test_get_observation_fields(self, **kwargs):
        """ get_observation_fields() work as expected (basic use)"""
        mock = kwargs['mock']

        mock.get('https://www.inaturalist.org/observation_fields.json?q=sex&page=2',
                 json=PAGE_2_JSON_RESPONSE, status_code=200)

        obs_fields = get_observation_fields(search_query="sex", page=2)
        assert  obs_fields == PAGE_2_JSON_RESPONSE


    @requests_mock.Mocker(kw='mock')
    def test_get_all_observation_fields(self, **kwargs):
        """get_all_observation_fields() is able to paginate, accepts a search query and return correct results"""
        mock = kwargs['mock']

        mock.get('https://www.inaturalist.org/observation_fields.json?q=sex&page=1',
                 json=PAGE_1_JSON_RESPONSE, status_code=200)

        mock.get('https://www.inaturalist.org/observation_fields.json?q=sex&page=2',
                 json=PAGE_2_JSON_RESPONSE, status_code=200)

        page_3_json_response = []
        mock.get('https://www.inaturalist.org/observation_fields.json?q=sex&page=3',
                 json=page_3_json_response, status_code=200)


        all_fields = get_all_observation_fields(search_query="sex")
        assert all_fields == PAGE_1_JSON_RESPONSE + PAGE_2_JSON_RESPONSE

    @requests_mock.Mocker(kw='mock')
    def test_get_all_observation_fields_noparam(self, **kwargs):
        """get_all_observation_fields() can also be called without a search query without errors"""

        mock = kwargs['mock']
        mock.get('https://www.inaturalist.org/observation_fields.json?page=1',
                 json=[], status_code=200)

        get_all_observation_fields()

    @requests_mock.Mocker(kw='mock')
    def test_get_access_token_fail(self, **kwargs):
        """ If we provide incorrect credentials to get_access_token(), an AuthenticationError is raised"""

        mock = kwargs['mock']
        rejection_json = {'error': 'invalid_client', 'error_description': 'Client authentication failed due to '
                                                                          'unknown client, no client authentication '
                                                                          'included, or unsupported authentication '
                                                                          'method.'}
        mock.post('https://www.inaturalist.org/oauth/token', json=rejection_json, status_code=401)

        with pytest.raises(AuthenticationError):
            get_access_token('username', 'password', 'app_id', 'app_secret')

    @requests_mock.Mocker(kw='mock')
    def test_get_access_token(self, **kwargs):
        """ Test a successful call to get_access_token() """

        mock = kwargs['mock']
        accepted_json = {'access_token': '604e5df329b98eecd22bb0a84f88b68a075a023ac437f2317b02f3a9ba414a08',
                         'token_type': 'Bearer', 'scope': 'write', 'created_at': 1539352135}
        mock.post('https://www.inaturalist.org/oauth/token', json=accepted_json, status_code=200)

        token = get_access_token('valid_username', 'valid_password', 'valid_app_id', 'valid_app_secret')

        assert token == '604e5df329b98eecd22bb0a84f88b68a075a023ac437f2317b02f3a9ba414a08'

    @requests_mock.Mocker(kw='mock')
    def test_update_observation(self, **kwargs):
        mock = kwargs['mock']
        mock.put('https://www.inaturalist.org/observations/17932425.json',
                 json=_load_sample_json('update_observation_result.json'),
                 status_code=200)

        p = {'ignore_photos': 1,
             'observation': {'description': 'updated description v2 !'}}
        r = update_observation(observation_id=17932425, params=p, access_token='valid token')

        # If all goes well we got a single element representing the updated observation, enclosed in a list.
        assert len(r) == 1
        assert r[0]['id'] == 17932425
        assert r[0]['description'] == 'updated description v2 !'

    @requests_mock.Mocker(kw='mock')
    def test_update_nonexistent_observation(self, **kwargs):
        """When we try to update a non-existent observation, iNat returns an error 410 with "obs does not longer exists". """
        mock = kwargs['mock']
        mock.put('https://www.inaturalist.org/observations/999999999.json',
                 json={"error":"Cette observation n’existe plus."},
                 status_code=410)

        p = {'ignore_photos': 1,
             'observation': {'description': 'updated description v2 !'}}

        with pytest.raises(HTTPError) as excinfo:
            update_observation(observation_id=999999999, params=p, access_token='valid token')
        assert excinfo.value.response.status_code == 410
        assert excinfo.value.response.json() == {"error":"Cette observation n’existe plus."}

    @requests_mock.Mocker(kw='mock')
    def test_update_observation_not_mine(self, **kwargs):
        """When we try to update the obs of another user, iNat returns an error 410 with "obs does not longer exists"."""
        mock = kwargs['mock']
        mock.put('https://www.inaturalist.org/observations/16227955.json',
                 json={"error": "Cette observation n’existe plus."},
                 status_code=410)

        p = {'ignore_photos': 1,
             'observation': {'description': 'updated description v2 !'}}

        with pytest.raises(HTTPError) as excinfo:
            update_observation(observation_id=16227955, params=p, access_token='valid token for another user')
        assert excinfo.value.response.status_code == 410
        assert excinfo.value.response.json() == {"error": "Cette observation n’existe plus."}

    @requests_mock.Mocker(kw='mock')
    def test_create_observation(self, **kwargs):
        mock = kwargs['mock']

        mock.post('https://www.inaturalist.org/observations.json',
                  json=_load_sample_json('create_observation_result.json'), status_code=200)


        params = {'observation':
                      {'species_guess': 'Pieris rapae'},
                  }

        r = create_observations(params=params, access_token='valid token')
        assert len(r) == 1  # We added a single one
        assert r[0]['latitude'] is None  # We have the field, but it's none since we didn't submitted anything
        assert r[0]['taxon_id'] == 55626 # Pieris Rapae @ iNaturalist

    @requests_mock.Mocker(kw='mock')
    def test_create_observation_fail(self, **kwargs):
        mock = kwargs['mock']
        params = {'observation':
                      {'species_guess': 'Pieris rapae',
                       # Some invalid data so the observation is rejected...
                       'observed_on_string': (datetime.now() + timedelta(days=1)).isoformat(),
                       'latitude': 200,
                       }
                  }

        mock.post('https://www.inaturalist.org/observations.json',
                  json=_load_sample_json('create_observation_fail.json'), status_code=422)

        with pytest.raises(HTTPError) as excinfo:
            create_observations(params=params, access_token='valid token')
        assert excinfo.value.response.status_code == 422
        assert 'errors' in excinfo.value.response.json()  # iNat also give details about the errors

    @requests_mock.Mocker(kw='mock')
    def test_put_observation_field_values(self, **kwargs):
        mock = kwargs['mock']
        mock.put('https://www.inaturalist.org/observation_field_values/31',
                 json=_load_sample_json('put_observation_field_value_result.json'),
                 status_code=200)

        r = put_observation_field_values(observation_id=18166477,
                                         observation_field_id=31,  # Animal behavior
                                         value='fouraging',
                                         access_token='valid token')

        assert r['id'] == 31
        assert r['observation_field_id'] == 31
        assert r['observation_id'] == 18166477
        assert r['value'] == 'fouraging'




# Code to access the (read-only, but fast) Node based public iNaturalist API
# See: http://api.inaturalist.org/v1/docs/
from time import sleep
from typing import Dict, Any, List

import requests
from urllib.parse import urljoin

from pyinaturalist.constants import THROTTLING_DELAY, INAT_NODE_API_BASE_URL
from pyinaturalist.exceptions import ObservationNotFound
from pyinaturalist.helpers import merge_two_dicts

PER_PAGE_RESULTS = 30  # Paginated queries: how many records do we ask per page?


def make_inaturalist_api_get_call(endpoint: str, params: Dict, **kwargs) -> requests.Response:
    """Make an API call to iNaturalist.

    endpoint is a string such as 'observations' !! do not put / in front
    method: 'GET', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE'
    kwargs are passed to requests.request
    Returns a requests.Response object
    """
    headers = {'Accept': 'application/json'}

    response = requests.get(urljoin(INAT_NODE_API_BASE_URL, endpoint), params, headers=headers, **kwargs)
    return response


def get_observation(observation_id: int) -> Dict[str, Any]:
    """Get details about an observation.

    :param observation_id:
    :returns: a dict with details on the observation
    :raises: ObservationNotFound
    """

    r = get_observations(params={'id': observation_id})
    if r['results']:
        return r['results'][0]

    raise ObservationNotFound()


def get_observations(params: Dict) -> Dict[str, Any]:
    """Search observations, see: http://api.inaturalist.org/v1/docs/#!/Observations/get_observations.

    Returns the parsed JSON returned by iNaturalist (observations in r['results'], a list of dicts)
    """

    r = make_inaturalist_api_get_call('observations', params=params)
    return r.json()


def get_all_observations(params: Dict) -> List[Dict[str, Any]]:
    """Like get_observations() but handles pagination so you get all the results in one shot.

    Some params will be overwritten: order_by, order, per_page, id_above (do NOT specify page when using this).

    Returns a list of dicts (one entry per observation)
    """

    # According to the doc: "The large size of the observations index prevents us from supporting the page parameter
    # when retrieving records from large result sets. If you need to retrieve large numbers of records, use the
    # per_page and id_above or id_below parameters instead.

    results = []  # type: List[Dict[str, Any]]
    id_above = 0

    while True:
        iteration_params = merge_two_dicts(params, {
            'order_by': 'id',
            'order': 'asc',
            'per_page': PER_PAGE_RESULTS,
            'id_above': id_above
        })

        page_obs = get_observations(params=iteration_params)
        results = results + page_obs['results']

        if page_obs['total_results'] <= PER_PAGE_RESULTS:
            return results

        sleep(THROTTLING_DELAY)
        id_above = results[-1]['id']



# Code used to access the (read/write, but slow) Rails based API of iNaturalist
# See: https://www.inaturalist.org/pages/api+reference
from time import sleep
from typing import Dict, Any, List, BinaryIO, Union  # noqa: F401

import requests

from pyinaturalist.constants import THROTTLING_DELAY, INAT_BASE_URL
from pyinaturalist.exceptions import AuthenticationError


def get_observation_fields(search_query: str="", page: int=1) -> List[Dict[str, Any]]:
    """
    Search the (globally available) observation
    :param search_query:
    :param page:
    :return:
    """
    payload = {
        'q': search_query,
        'page': page
    }  # type: Dict[str, Union[int, str]]

    response = requests.get("{base_url}/observation_fields.json".format(base_url=INAT_BASE_URL), params=payload)
    return response.json()


def get_all_observation_fields(search_query: str="") -> List[Dict[str, Any]]:
    """
    Like get_observation_fields(), but handles pagination for you.

    :param search_query: a string to search
    """
    results = []  # type: List[Dict[str, Any]]
    page = 1

    while True:
        r = get_observation_fields(search_query=search_query, page=page)

        if not r:
            return results

        results = results + r
        page = page + 1
        sleep(THROTTLING_DELAY)


def put_observation_field_values(observation_id: int, observation_field_id: int, value: Any,
                                 access_token: str) -> Dict[str, Any]:
    # TODO: Also implement a put_or_update_observation_field_values() that deletes then recreates the field_value?
    # TODO: Write example use in docstring.
    # TODO: Return some meaningful exception if it fails because the field is already set.
    # TODO: Also show in example  to obtain the observation_field_id?
    # TODO: What happens when parameters are invalid
    # TODO: It appears pushing the same value/pair twice in a row (but deleting it meanwhile via the UI)...
    # TODO: ...triggers an error 404 the second time (report to iNaturalist?)
    """Sets an observation field (value) on an observation.

    :param observation_id:
    :param observation_field_id:
    :param value
    :param access_token: access_token: the access token, as returned by :func:`get_access_token()`

    :returns: iNaturalist's response as a dict, for example:
            {'id': 31,
             'observation_id': 18166477,
             'observation_field_id': 31,
             'value': 'fouraging',
             'created_at': '2012-09-29T11:05:44.935+02:00',
             'updated_at': '2018-11-13T10:49:47.985+01:00',
             'user_id': 1,
             'updater_id': 1263313,
             'uuid': 'b404b654-1bf0-4299-9288-52eeda7ac0db',
             'created_at_utc': '2012-09-29T09:05:44.935Z',
             'updated_at_utc': '2018-11-13T09:49:47.985Z'}

    Will fail if this observation_field is already set for this observation.
    """

    payload = {
        'observation_field_value': {
            'observation_id': observation_id,
            'observation_field_id': observation_field_id,
            'value': value
        }
    }

    response = requests.put("{base_url}/observation_field_values/{id}".format(base_url=INAT_BASE_URL,
                                                                              id=observation_field_id),
                            headers=_build_auth_header(access_token),
                            json=payload)

    response.raise_for_status()
    return response.json()


def get_access_token(username: str, password: str, app_id: str, app_secret: str) -> str:
    """
    Get an access token using the user's iNaturalist username and password.

    (you still need an iNaturalist app to do this)

    :param username:
    :param password:
    :param app_id:
    :param app_secret:
    :return: the access token, example use: headers = {"Authorization": "Bearer %s" % access_token}
    """
    payload = {
        'client_id': app_id,
        'client_secret': app_secret,
        'grant_type': "password",
        'username': username,
        'password': password
    }

    response = requests.post("{base_url}/oauth/token".format(base_url=INAT_BASE_URL), payload)
    try:
        return response.json()["access_token"]
    except KeyError:
        raise AuthenticationError("Authentication error, please check credentials.")


def _build_auth_header(access_token: str) -> Dict[str, str]:
    return {"Authorization": "Bearer %s" % access_token}


def add_photo_to_observation(observation_id: int, file_object: BinaryIO, access_token: str):
    """Upload a picture and assign it to an existing observation.

    :param observation_id: the ID of the observation
    :param file_object: a file-like object for the picture. Example: open('/Users/nicolasnoe/vespa.jpg', 'rb')
    :param access_token: the access token, as returned by :func:`get_access_token()`
    """
    data = {'observation_photo[observation_id]': observation_id}
    file_data = {'file': file_object}

    response = requests.post(url="{base_url}/observation_photos".format(base_url=INAT_BASE_URL),
                             headers=_build_auth_header(access_token),
                             data=data,
                             files=file_data)

    return response.json()


def create_observations(params: Dict[str, Dict[str, Any]], access_token: str) -> List[Dict[str, Any]]:
    """Create a single or several (if passed an array) observations).

    :param params:
    :param access_token: the access token, as returned by :func:`get_access_token()`

    :return: iNaturalist's JSON response, as a Python object
    :raise: requests.HTTPError, if the call is not successful. iNaturalist returns an error 422 (unprocessable entity)
            if it rejects the observation data (for example an observation date in the future or a latitude > 90. In
            that case the exception's `response` attribute give details about the errors.

    allowed params: see https://www.inaturalist.org/pages/api+reference#post-observations

    Example:

        params = {'observation':
            {'species_guess': 'Pieris rapae'},
        }

    TODO investigate: according to the doc, we should be able to pass multiple observations (in an array, and in
    renaming observation to observations, but as far as I saw they are not created (while a status of 200 is returned)
    """
    response = requests.post(url="{base_url}/observations.json".format(base_url=INAT_BASE_URL),
                             json=params,
                             headers=_build_auth_header(access_token))
    response.raise_for_status()
    return response.json()


def update_observation(observation_id: int, params: Dict[str, Any], access_token: str) -> List[Dict[str, Any]]:
    """
    Update a single observation. See https://www.inaturalist.org/pages/api+reference#put-observations-id

    :param observation_id: the ID of the observation to update
    :param params: to be passed to iNaturalist API
    :param access_token: the access token, as returned by :func:`get_access_token()`

    :return: iNaturalist's JSON response, as a Python object
    :raise: requests.HTTPError, if the call is not successful. iNaturalist returns an error 410 if the observation
            doesn't exists or belongs to another user (as of November 2018).
    """

    response = requests.put(url="{base_url}/observations/{id}.json".format(base_url=INAT_BASE_URL, id=observation_id),
                            json=params,
                            headers=_build_auth_header(access_token))
    response.raise_for_status()
    return response.json()


# TODO: test this
# TODO: document this
def delete_observation(observation_id: int, access_token: str) -> List[Dict[str, Any]]:
    """
    Delete an observation.

    :param observation_id:
    :param access_token:
    :return:
    """

    response = requests.delete(url="{base_url}/observations/{id}.json".format(base_url=INAT_BASE_URL,
                                                                              id=observation_id),
                               headers=_build_auth_header(access_token))
    response.raise_for_status()
    return response.json()



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()

history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='pyinaturalist',
    version='0.6.0',
    description='Python client for the iNaturalist APIs',
    long_description=readme + '\n\n' + history,
    author='Nicolas Noé',
    author_email='nicolas@niconoe.eu',
    url='https://github.com/inbo/pyinaturalist',
    packages=[
        'pyinaturalist',
    ],
    package_dir={'pyinaturalist': 'pyinaturalist'},
    include_package_data=True,
    install_requires=[
        'requests>=2.20.0'
    ],
    license='MIT',
    zip_safe=False,
    keywords='pyinaturalist',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
